/*
 * Copyright (C) 1998-2003 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
/**
 ** sickCntrlTaskCodels.c
 **
 ** Codels used by the control task sickCntrlTask
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#ifndef UNIX
#include <vxWorks.h>
#else
#include <portLib.h>
#endif
#include <stdio.h>

/* sick module header */
#include "sickNHeader.h"


/*------------------------------------------------------------------------
 * sickContinuousShotCntrl  -  control codel of CONTROL request ContinuousShot
 *
 * Description:    
 * Report (bilan): OK 
 *
 * Return:    OK ou ERROR
 */

STATUS sick_continuous_shot_cntrl (SICK_SCAN_MODES *scan_mode, int *bilan)
{
  SDI_F->required_control_mode = *scan_mode == SICK_PERIODIC_SCAN ?
				 SICK_SCAN_CONTINUOUSLY_ON_RQST_MODE : 
				 SICK_SCAN_CONTINUOUSLY_MODE;

  return OK;
}

/*------------------------------------------------------------------------
 * sickSetSegmentationParamsCntrl  -  control codel of CONTROL request SetSegmentationParams
 *
 * Description:    
 * Report (bilan): OK 
 *                 S_sick_BAD_PARAMS
 *
 * Return:    OK ou ERROR
 */

STATUS sickSetSegmentationParamsCntrl (SICK_SEG_PARAMS *segmentationParams, 
				       int *bilan)
{

#ifndef SEG_VARIANCES
  if (segmentationParams->getVariances) {
    printf ("sick error : this module must be compiled with SEG_VARIANCES\n"
	    "                to get variances on segments\n");
    *bilan = S_sick_BAD_PARAMS;
    return ERROR;
  }
#endif 

  return OK;
}

/*------------------------------------------------------------------------
 * sickSetEnvFileCntrl  -  control codel of CONTROL request SetEnvFileName
 *
 * Description:    
 * Report (bilan): OK 
 *                 S_sick_CANNOT_OPEN_FILE
 *
 * Return:    OK ou ERROR
 */

STATUS sickSetEnvFileCntrl (GEN_STRING256 *envFileName, int *bilan)
{
  FILE *pfile;
  if (!(pfile  = fopen(envFileName->name, "r"))) {
    printf ("sickSetEnvFileExec: cannot open file %s\n", envFileName->name);
    *bilan = S_sick_CANNOT_OPEN_FILE;
    return ERROR;
  }
  fclose (pfile);
  return OK;
}

/*------------------------------------------------------------------------
 * ReadSickFiles
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_sick_FILE_NOT_FOUND
 */

/* sickReadSickFiles  -  codel EXEC de ReadSickFiles
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
STATUS sickSetSickDataFilesCntrl (SICK_SAVE_FILE_PARAMS *saveFileParams, 
				  int *bilan)
{
  char fileName[64];
  FILE *pfile;

  sprintf(fileName, "%s.%03d", 
	  saveFileParams->baseName, saveFileParams->num);  
  if (!(pfile=fopen(fileName, "r"))) {
    printf ("sickSetReadSickFilesCntrl : cannot read %s\n", fileName);
    *bilan = S_sick_FILE_NOT_FOUND;
    SDI_F->readSickFile = SICK_FALSE;
    return ERROR;
  }
  fclose(pfile);
  SDI_F->readSickFile = SICK_TRUE;
  return OK;
}

