/**
 ** sickScanCodels.c
 **
 ** Codels appeles par la tache sickScan
 **
 ** Auteur: 
 ** Date: 
 **
 **/

#ifndef UNIX
#include <vxWorks.h>
#include <sysLib.h>
#else
#include <portLib.h>
#include <math.h>
#endif
#include <stdio.h>

/* sick module header */
#include "sickNHeader.h"

#include "h2mathLib.h"
#include "h2timeLib.h"

#include "sickConst.h"
#include "sick_func_proto.h"
#include "sick_local_str.h"
/*#include "sick_save_proto.h"*/
#include "sick_segs.h"
#include "sick_save.h"

/*------------------------------------------------------------------------
 *
 * sick_init  --  Codel d'initialisation (SDI/f, ...)
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS sick_init(int *bilan)
{
  /* init */
  return sick_func_init (bilan);
}

/*------------------------------------------------------------------------
 *
 * sick_module_end  --  Codel de terminaison
 *
 * Description: 
 * 
 * Retourne:    OK ou ERROR
 */

STATUS sick_module_end(void)
{
  sick_func_end();
  return OK;
}

/*------------------------------------------------------------------------
 *
 * sick_scan_func  --  codel de l'activite' permanente
 *
 * Description: 
 * 
 * Bilans:      OK 
 *              S_sick_sick_fatal_error
 * 
 * Retourne:    OK ou ERROR
 */

STATUS sick_scan_func(int *bilan)
{
  SICK_SEGS_POSTER_STR acqSegs;
  int measure=FALSE;
  static int waitData = 0;
  static int prevNoPoint = TRUE;

  /* Check State */
  if (SDI_F->state != SICK_CONFIGURATED_STATE) {
    /* check mode */
    if (SDI_F->control_mode != SICK_CONFIGURATION_MODE) {
      printf ("sick_scan_func: sick not in SICK_CONFIGURATED_STATE (=%d) and not in SICK_CONFIGURATION_MODE (=%d)\n"
	      " state = %d control_mode = %d\n",
	      SICK_CONFIGURATED_STATE, SICK_CONFIGURATION_MODE,
	      SDI_F->state, SDI_F->control_mode);
      *bilan = S_sick_fatal_error;
      return ERROR;
    }
    /* ok */
    return OK;
  }

  /* switch operation mode */
  switch (SDI_F->control_mode) {
  case SICK_CONFIGURATION_MODE:
    break;


  case SICK_SCAN_CONTINUOUSLY_ON_RQST_MODE:
    if (!waitData) {
      if (sick_ask_for_measures (bilan) != OK) {
	sick_reset_inter (bilan);
	SDI_F->control_mode = SICK_CONFIGURATION_MODE;
      }
      else {
	sick_flush_serial (bilan);
	waitData = 1;
      }
    }

    /* ask for measures */
  case SICK_SCAN_ON_RQST_MODE:

    /* read measures */
  case SICK_SCAN_CONTINUOUSLY_MODE:

    /* Simul without env file */
    if (SDI_F->initParams.acqWay == SICK_ACQ_SIMUL_ENV && !SDI_F->readEnvFile)
      return OK;

    /* get data (real or simul) */
    if (SDI_F->initParams.acqWay != SICK_ACQ_REPLAY_FILES) {
      measure = sick_wait_measures_tram (bilan);
      if (*bilan != OK && *bilan != S_sick_cannot_connect_to_pom) {
	SDI_F->control_mode = SICK_CONFIGURATION_MODE;
	SDI_F->state = SICK_FREE_STATE;
	return OK;
      }
    }
    
    /* read measure by file */
    else {
      if (SDI_F->readSickFile) {
	if (!(measure = sick_read_measure_in_file())) {
	  SDI_F->readSickFile = SICK_FALSE;
	  return OK;
	}
      }
    }

    /* put measures in posters (and computes cartesian points) */
    if (measure) {
      waitData = 0;
      if (SDI_F->polarMeasures.measures.header.np == 0) {
	if (prevNoPoint == FALSE) {
	  prevNoPoint = TRUE;
	  printf("sick : no points !\n");
	}
	measure = FALSE;
      }
      else {
	if (prevNoPoint) 
	  printf ("sick: %d points : OK\n", SDI_F->polarMeasures.measures.header.np);
	prevNoPoint = FALSE;
      }
      if (sick_update_measures (bilan) != OK) {
	SDI_F->state = SICK_FREE_STATE;
	return OK;
      }
    }

    /* segmentation */
    if (measure && SDI_F->initParams.exportData.exportSegments) {
      if (!sickSegmentation(&SDI_F->polarMeasures.measures.header, SDI_F->polarMeasures.measures.rho, 
			    &SDI_F->segParams, &acqSegs,SICK_SEGS_POSTER_ID,
			    SDI_F->initParams.exportData.cartesianFrame)) {
	printf ("sick : segmentation failed -> off \n"
		"       (rqst DataProduction to restart)\n");
/* XXX	SDI_F->initParams.exportData.exportSegments = SICK_FALSE; */
      }
    }

#ifdef DEBUG
    fprintf(stdout,"Export segmens in frame : %d\n",SDI_F->initParams.exportData.cartesianFrame);
    fprintf(stdout, "Cartesian Segments Measures\n");
    fprintf(stdout, "Robot Position : %f %f %f\n",
	    SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.x,
	    SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.y,
	    SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.z);
    
    int i=0;
    for (i=0;i < acqSegs.Header.nbSegs;i++) {
      fprintf(stdout,"i=%d (%lf,%lf) (%lf,%lf)\n",i,
	      acqSegs.Segs.segs[i].x1,
	      acqSegs.Segs.segs[i].y1,
	      acqSegs.Segs.segs[i].x2,
	      acqSegs.Segs.segs[i].y2);      
    }
#endif    
    break;
  } /* switch */

  /* ok */
  return OK;
}

/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* InitExec  -  codel EXEC de Init
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_init_exec (SICK_INIT_PARAMS *init, int *bilan)
{
  return sick_func_init_exec(init, bilan);
}

/*------------------------------------------------------------------------
 * Reset
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_sick_serial_input_error
 *              S_sick_serial_output_error
 *              S_sick_reset_error
 */

/* sick_reset_start  -  codel START de Reset
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_reset_start (int *bilan)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;
  int baud;

  /* find baud rate */
  if ((baud = sick_find_baud_rate (&SDI_F->sick_config.baud, bilan)) < 0) {
    fprintf (stderr, "\n\n\tSick must be reset!!\n");
    return sick_reset_inter (bilan);
  }

  /* flush serial port */
  if (sick_flush_serial (bilan) != OK)
    return sick_reset_inter (bilan);

  /* compose parser commands */
  sick_compose_parser_cmds_list (TRUE);

  /* set global mode */
  SDI_F->control_mode = SICK_CONFIGURATION_MODE;
  SDI_F->state = parser->state = parser->cmds[0];
  parser->time = 0;

  return EXEC;
}

/* sick_reset  -  codel EXEC de Reset
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_reset (int *bilan)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;
  int event;

  /* when calling from one shoot */
  if (parser->index >= parser->ncmd) 
    return ETHER;

  /* call dynamic config */
  event = sick_dynamic_configurations (bilan);

  /* error */
  if (event == FAIL)
    return  sick_reset_inter (bilan);

  /* ETHER */
  if (event == ETHER)
    return END;

  /* end parser state */
  if (event == END) {
    /* step */
    parser->index++;
    parser->time = 0;

    /* there is more cmds */
    if (parser->index < parser->ncmd)
      parser->state = parser->cmds[parser->index];
    else
      return END;
  }
  
  parser->time += SICK_SCAN_TASK_PERIOD;

  if (parser->time > SICK_REPONSE_TIME_MAX) {
    *bilan = S_sick_time_out;
    return sick_reset_inter (bilan);
  }
  /* EXEC */
  return EXEC;
}

/* sick_reset_end  -  codel END de Reset
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_reset_end (int *bilan)
{
  /* configurated */
  SDI_F->state = SICK_CONFIGURATED_STATE;
  SDI_F->control_mode = SDI_F->sick_config.control_mode;
  
  return ETHER;
}

/* sick_reset_inter  -  codel INTER de Reset
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_reset_inter (int *bilan)
{
  /* init of sick config */
  SDI_F->state = SICK_FREE_STATE;
  
  return ETHER;
}

/*------------------------------------------------------------------------
 * Configuration
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* sick_configuration_start  -  codel START de Eval
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_configuration_start (SICK_USER_STR *user,
					 SICK_CONFIG_STR *(sick_config), 
					 int *bilan)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;


  /* check state */
  if (SDI_F->state == SICK_FREE_STATE) {
    *bilan = S_sick_isnot_initialized;
    return ETHER;
  }

  /* flush serial port */
  if (sick_flush_serial (bilan) != OK)
    return sick_reset_inter (bilan);

  /* compose parser commands */
  if (!sick_compose_parser_cmds_list (FALSE)) 
    return ETHER;

  /* set global mode */
  SDI_F->control_mode = SICK_CONFIGURATION_MODE;
  SDI_F->state = parser->state = parser->cmds[0];
  parser->time = 0;

  return EXEC;

}

/* sick_eval  -  codel EXEC de Eval
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_configuration_exec (SICK_USER_STR *user,
					SICK_CONFIG_STR *(sick_config), int *bilan)
{
  return sick_reset (bilan);
}

/* sick_configuration_end  -  codel END de Eval
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_configuration_end (SICK_USER_STR *user,
				       SICK_CONFIG_STR *(sick_config), int *bilan)
{
  return sick_reset_end (bilan);
}

/* sick_configuration_inter  -  codel INTER de Eval
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_configuration_inter (SICK_USER_STR *user,
					 SICK_CONFIG_STR *(sick_config), int *bilan)
{
  /* XXX ? (can be used !) */
  return sick_reset_inter (bilan);
}


/*------------------------------------------------------------------------
 * Continuous_Shot
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* sick_continuous_shot_start  -  codel START de Eval
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_continuous_shot_start (SICK_SCAN_MODES *user,
					 int *bilan)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;


  /* check state */
  if (SDI_F->state == SICK_FREE_STATE) {
    *bilan = S_sick_isnot_initialized;
    return ETHER;
  }

  /* flush serial port */
  if (sick_flush_serial (bilan) != OK)
    return sick_reset_inter (bilan);

  /* compose parser commands */
  if (!sick_compose_parser_cmds_list (FALSE)) 
    return ETHER;

  /* set global mode */
  SDI_F->control_mode = SICK_CONFIGURATION_MODE;
  SDI_F->state = parser->state = parser->cmds[0];
  parser->time = 0;

  return EXEC;

}

/* sick_eval  -  codel EXEC de Eval
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_continuous_shot_exec (SICK_SCAN_MODES *user,
					  int *bilan)
{
  return sick_reset (bilan);
}

/* sick_continuous_shot_end  -  codel END de Eval
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_continuous_shot_end (SICK_SCAN_MODES *user,
					 int *bilan)
{
  return sick_reset_end (bilan);
}

/* sick_continuous_shot_inter  -  codel INTER de Eval
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_continuous_shot_inter (SICK_SCAN_MODES *user,
					   int *bilan)
{
  /* XXX ? (can be used !) */
  return sick_reset_inter (bilan);
}


/*------------------------------------------------------------------------
 * OneShoot
 *
 * Description: 
 *
 * Bilans:      OK 
 */

/* sick_one_shoot_start  -  codel START de OneShoot
   Retourne:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_one_shoot_start (int *bilan)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;

  /* check state */
  if (SDI_F->state == SICK_FREE_STATE) {
    *bilan = S_sick_isnot_initialized;
    return ETHER;
  }

  /* check current global mode */
  if (SDI_F->control_mode == SICK_SCAN_ON_RQST_MODE) {
    if (sick_ask_for_measures (bilan) != OK)
      return sick_reset_inter (bilan);
    else
      return ETHER;
  }

  /* flush serial port */
  if (sick_flush_serial (bilan) != OK)
    return sick_reset_inter (bilan);

  /* init of user config */
  SDI_F->required_control_mode = SICK_SCAN_ON_RQST_MODE;

  /* compose parser commands */
  if (!sick_compose_parser_cmds_list (FALSE)) 
    return ETHER;

  /* set global mode */
  SDI_F->control_mode = SICK_CONFIGURATION_MODE;
  SDI_F->state = parser->state = parser->cmds[0];
  parser->time = 0;

  return EXEC;
}

/* sick_one_shoot  -  codel EXEC de OneShoot
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_one_shoot_exec (int *bilan)
{
  ACTIVITY_EVENT event;

  /* configurated it */
  event = sick_reset (bilan);

  /* exec */
  if (event == EXEC)
    return EXEC;

  /* end */
  if (event == END) {
    if (sick_ask_for_measures (bilan) != OK)
      return sick_reset_inter (bilan);
    else
      return sick_reset_end (bilan);
  }

  return ETHER;
}

/* sick_one_shoot_end  -  codel END de OneShoot
   Retourne:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_one_shoot_end (int *bilan)
{
  /* XXX ? (however, never used) */
    return sick_reset_end (bilan);
}

/* sick_one_shoot_inter  -  codel INTER de OneShoot
   Retourne:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sick_one_shoot_inter (int *bilan)
{
  /* XXX ? (can be used !) */
    return sick_reset_inter (bilan);
}

/*------------------------------------------------------------------------
 * Save
 *
 * Description: 
 *
 * Bilans:      OK 
 *              S_sick_CANNOT_OPEN_FILE
 */

/* sickSaveExec  -  codel EXEC de Save
   Retourne:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT sickSaveExec (SICK_SAVE_FILE_PARAMS *saveFileParams, int *bilan)
{
  char fileName[SICK_MAX_STRLEN+8];
  static unsigned int prevTime=0;
  unsigned long int time;
  static POM_EULER prevRobot={1000,1000,1000,1000,1000,1000};
  POM_EULER robot;
  static int count=0;
  

  robot.x=SDI_F->polarMeasures.measures.header.sickPomPos.mainToBase.euler.x;
  robot.y=SDI_F->polarMeasures.measures.header.sickPomPos.mainToBase.euler.y;
  robot.yaw=SDI_F->polarMeasures.measures.header.sickPomPos.mainToBase.euler.yaw;

  /* Attend des donnes fraches */
  if (saveFileParams->saveMode != SICK_SAVE_ONCE) {

    /* Donnes mises a jour ? */
    time = SDI_F->polarMeasures.measures.header.rcv_time;
    if (time == prevTime)
      return EXEC;
    prevTime = time;

    /* Compteur */
    if (saveFileParams->saveContinuousFrequency > 1) {
      if (count>0) {
	count--;
	return EXEC;
      }
      count = saveFileParams->saveContinuousFrequency;
    }

    /* Attend deplacement du robot */
    if (saveFileParams->saveMode == SICK_SAVE_CONTINUOUS_MOVE) {
      if (fabs(robot.x - prevRobot.x) < 0.001 &&
	  fabs(robot.y - prevRobot.y) < 0.001 &&
	  fabs(robot.yaw - prevRobot.yaw) < 0.001)
	return EXEC;
      prevRobot = robot;
    }
  }

  /* Nom fichier */
  sprintf(fileName, "%s.%05d", 
	  SDI_F->saveFileParams.baseName, SDI_F->saveFileParams.num);
  
  /* Sauvegarde */
  if (!sickSaveFile(fileName, &SDI_F->sick_config, 
		    &SDI_F->polarMeasures.measures.header, SDI_F->polarMeasures.measures.rho)) {
    printf("sickSaveExec: cannot open %s\n", fileName);
    *bilan = S_sick_CANNOT_OPEN_FILE;
    return ETHER;
  }
  (SDI_F->saveFileParams.num)++;

  /* Une seule acquisition */
  if (saveFileParams->saveMode == SICK_SAVE_ONCE)
    return ETHER;

  /* Acquisition en boucle */
  return EXEC;
}

/*------------------------------------------------------------------------
 * Log
 *
 * Description: 
 *
 * Reports:      OK
 *              S_sick_CANNOT_OPEN_FILE
 *              S_sick_CANNOT_WRITE_FILE
 */

static FILE *sickLogFile;

/* sickLogStart  -  codel START of Log
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sickLogStart(GEN_STRING128 *logFileName, int *report)
{
  if (!(sickLogFile = fopen(logFileName->name, "w"))) {
    printf ("[sick-genom] cannot open odometer file %s\n", logFileName->name);
    *report = S_sick_CANNOT_OPEN_FILE;
    return ETHER;
  }

  printf ("sickLogStart: log file %s found\n", logFileName->name);

  return EXEC;
}

/* sickLogExec  -  codel EXEC of Log
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sickLogExec(GEN_STRING128 *logFileName, int *report)
{
  H2TIMESPEC tv;

  /* get date */
  h2GetTimeSpec(&tv);

  /* save date */
  if (!fprintf(sickLogFile, "%lf\n", 
	       tv.tv_sec + tv.tv_nsec*1.0E-9)) {
    printf ("[sick-genom] cannot log in %s\n", logFileName->name);
    sickLogFile = 0;
    *report = S_sick_CANNOT_WRITE_FILE;
    return ETHER;
  }

  fprintf(sickLogFile, "Robot Position : %f %f %f\n",
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.x,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.y,
	  SDI_F->cartMeasures.header.sickPomPos.mainToOrigin.euler.z);
  fprintf(sickLogFile, "Polar Measures\n");
  int i;
  for (;i < SDI_F->polarMeasures.measures.header.np;i++) {
    fprintf(sickLogFile,"i=%d rho=%lf theta=%lf\n",i,SDI_F->polarMeasures.measures.rho[i],SDI_F->polarMeasures.theta[i]);
  }

  return EXEC;  
}

/* sickLogInter  -  codel END of Log
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
sickLogInter(GEN_STRING128 *logFileName, int *report)
{
  fclose(sickLogFile);
  sickLogFile = NULL;

  return ETHER;
}


