/*  
 * Copyright (c) 1998-2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

#include <h2mathLib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef VXWORKS
#include <vxWorks.h>
#include <sysLib.h>
#include <logLib.h>
#define M_PI PI
#else
#include <portLib.h>
#endif /* VXWORKS */

#include <taskLib.h>
#include <errnoLib.h>

/* sick module header */
#include "sickNHeader.h"

#include "sickConst.h"

/* sick_lib */
#include "sick_lib.h"


/* local */
#include "sick_local_str.h"

#ifdef READ_SIMULATED_POS
#include "genPosStruct.h"
#include "auto/locoPosterLib.h"
#endif

#include "sick_convert.h"

/*
 * protos
 */
#include "libconvert/sick_save.h"

/* sick lib object */
static SICK_LIB_STR *sick_id = NULL;

/* reset commands list */
static SICK_SCAN_TASK_STATE reset_cmds[] = SICK_RESET_CMDS_LIST;

/* current sick pos */
static POSTER_ID sickPosPosterId;
typedef enum SICK_POS_POSTER_TYPE {
  SICK_NO_POS_POSTER,
  SICK_POM_SENSOR_POS_POSTER,
  SICK_POM_POS_POSTER,
  SICK_POM_MEPOS_POSTER
} SICK_POS_POSTER_TYPE;
static SICK_POS_POSTER_TYPE sickPosPosterType;


/* ------------------------ PROTOTYPES ------------------------- */
static BOOL sickPosterFind(void);

extern STATUS sick_ask_for_measures ( int *bilan );
extern int sick_compose_parser_cmds_list ( int reset );
extern ACTIVITY_EVENT sick_dynamic_configurations ( int *bilan );


static void sick_update_global_sick_pos(POM_SENSOR_POS *robot,
					SICK_LIB_STR *sick_id);
static STATUS sick_get_robot_position (POM_SENSOR_POS *pos,int *bilan);
STATUS sick_func_init (int *bilan);
static void sick_init_data ();

extern STATUS sick_func_end ( void );
extern STATUS sick_func_init ( int *bilan );
extern ACTIVITY_EVENT sick_func_init_exec ( SICK_INIT_PARAMS *init, int *bilan );
static STATUS sick_first_configurations ( int *bilan );
static void sick_set_length_resolution ( void );
static void sick_set_scanning_params ( void );
static void sick_update_sick_data_str ( SICK_SCAN_TASK_STATE state );
extern STATUS sick_flush_serial ( int *bilan );
BOOL sick_wait_measures_tram (int *bilan) ;
BOOL sick_func_read_file_env(char *file);
extern BOOL sick_read_measure_in_file ( void );
int sick_find_baud_rate (SICK_BAUD_RATE *br, int *bilan);
static void sick_set_baud_rate ( void );

static void sick_choose_robot_pos (int n, POM_SENSOR_POS *p);
static STATUS sick_serial_change_baud (int baud, int *bilan);
static void sick_update_sick_data_str (SICK_SCAN_TASK_STATE state);

/* ---------------------------------------------------------------------- */


#include "pomStruct.h"
/*
 * Init of sick functions
 */

STATUS
sick_func_init (int *bilan)
{
  /* init data */
  sick_init_data ();

  if (sickScanPolarHeaderPosterWrite (SICK_SCANPOLAR_POSTER_ID, 
				      &SDI_F->polarMeasures.measures.header) != OK) {

    printf ("sick_func_init: sickScanPolarHeaderPosterWrite failed\n");
    *bilan = errnoGet ();
    return ERROR;
  }
  if (sickScanCartHeaderPosterWrite (SICK_SCANCART_POSTER_ID, 
				     &SDI_F->polarMeasures.measures.header) != OK) {
    printf ("sick_func_init: sickScanCartHeaderPosterWrite failed\n");
    *bilan = errnoGet ();
    return ERROR;
  }

#ifdef XR4000
#define DELAY (0.260) /* experimental value */
  SDI_F->synchroPosTunning = (int)((DELAY)/(EXEC_TASK_PERIOD(SICK_SCANTASK_NUM)));
#else
  SDI_F->synchroPosTunning = 0;
#endif

  printf ("SICK SYNCHRO TUNNING = %d\n", SDI_F->synchroPosTunning);

#ifdef SICK_MIN_ANGLE
  printf ("Sick Min angle set to %f\n", RAD_TO_DEG(SICK_MIN_ANGLE));
#endif
#ifdef SICK_MAX_ANGLE
  printf ("Sick Max angle set to %f\n", RAD_TO_DEG(SICK_MAX_ANGLE));
#endif

  /* ok */
  return OK;
}

/*
 * End of sick functions
 */

STATUS
sick_func_end (void)
{
  sick_lib_end(sick_id);
  sick_id = NULL;
  return OK;
}

/* ----------------------------------------------------------------------
 * Init request
 */

ACTIVITY_EVENT sick_func_init_exec (SICK_INIT_PARAMS *init, int *bilan)
{
  POM_SENSOR_POS pos;

  /* get data by file */
  if (init->acqWay == SICK_ACQ_REPLAY_FILES) {
    SDI_F->state = SICK_CONFIGURATED_STATE;
    SDI_F->readSickFile = SICK_TRUE;
    SDI_F->control_mode = SICK_SCAN_ON_RQST_MODE;
    return ETHER;
  }

  /* for simulation */
  SDI_F->readEnvFile = SICK_FALSE;
  if (init->acqWay == SICK_ACQ_SIMUL_ENV && SDI_F->envFileName.name[0] != '\0') {
    if (sick_func_read_file_env(SDI_F->envFileName.name))
      SDI_F->readEnvFile = SICK_TRUE;
  }

  /* real sick */
  /* init of sick */
  if (!sick_id) {
    sick_id = sick_lib_init (SDI_F->initParams.serialPort.name,
			     SICK_SERIAL_DEFAULT_RATE);
    
    /* check */
    if (!sick_id) {
      *bilan = errnoGet ();
      return ETHER;
    }

    /* wait serial tasks */
    taskDelay (sysClkRateGet() * 2);
  }

  /* find baud rate */
  if (sick_find_baud_rate (&SDI_F->sick_config.baud, bilan) < 0) {
    fprintf (stderr, "\n\n\tSick must be reset!!\n");
    return ETHER;
  }

  /* configurate sick */
  sick_first_configurations (bilan);

  /* robot pos */
  if (sick_get_robot_position (&pos,bilan) != OK) {
    *bilan = S_sick_cannot_connect_to_pom;
    fprintf(stderr,"sick: cannot read robot pos\n");
    return ETHER;
  }

  return ETHER;
}

/* ------------------------------------------------------------
 *
 * sickPosterFind
 *
 */
static BOOL sickPosterFind()
{
  int status;
  sickPosPosterType = SICK_NO_POS_POSTER;

  printf ("sickReadRobotPosition:  poster %s\n",
	  SDI_F->initParams.sickPosPosterName.name);
  
  /* look for a pom poster of type POM_POS or POM_ME_POS */
  if (SDI_F->initParams.sickPosPosterName.name[0] == '\0') {
    printf ("sickReadRobotPosition:  NO POSTER POS\n");
    return FALSE;
  }

  /* try POM_SENSOR_POS */
#ifdef SICK2
  status = sick2POM_SENSOR_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
					 &sickPosPosterId);
#else
  status = sickPOM_SENSOR_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
					&sickPosPosterId);
#endif
  if (status == OK) {
    printf ("sick: POM_SENSOR_POS poster\n");
    sickPosPosterType = SICK_POM_SENSOR_POS_POSTER;
    return TRUE;
  }
  if(!sickPosPosterId) {
    printf ("sick:  cannot find poster %s\n",
	    SDI_F->initParams.sickPosPosterName.name);
    return FALSE;
  }

  /* try POM_POS */
#ifdef SICK2
  status = sick2POM_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
				  &sickPosPosterId);
#else
  status = sickPOM_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
				 &sickPosPosterId);
#endif
  if (status == OK) {
    printf ("sick: POM_POS poster. Give the position of the sick using request SetSickPosOnRobot\n");
    sickPosPosterType = SICK_POM_POS_POSTER;
    return TRUE;
  }

  /* try POM_ME_POS */
#ifdef SICK2
  status = sick2POM_ME_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
					 &sickPosPosterId);
#else
  status = sickPOM_ME_POSPosterFind(SDI_F->initParams.sickPosPosterName.name, 
					&sickPosPosterId);
#endif
  if (status == OK) {
    printf ("sick: POM_ME_POS poster. Give the position of the sick using request SetSickPosOnRobot\n");
    sickPosPosterType = SICK_POM_MEPOS_POSTER;
    return TRUE;
  }
  printf ("sick:  unvalid format for poster %s\n",
	  SDI_F->initParams.sickPosPosterName.name);
  return FALSE;
}

/* ----------------------------------------------------------------------
 * Sick Init data
 */
static void 
sick_init_data ()
{
  /* get pointer of sick and user config str */
  SICK_CONFIG_STR *config = &SDI_F->sick_config;
  SICK_USER_STR *user = &SDI_F->user_config;

  /* sick measures */
  SICK_MEASURES_STR *m = &SDI_F->polarMeasures.measures;


  /* init of sick config */
  SDI_F->state = SICK_FREE_STATE;
  /* sick global mode */
  SDI_F->control_mode = SICK_CONFIGURATION_MODE;
  
  /* sick config */
#ifdef SICK_NOT_CONTROLLABLE
  config->control_mode = SICK_SCAN_CONTINUOUSLY_MODE;
  config->angle_range = SICK_SCAN_180_DEG;
  config->angle_resolution = SICK_SCAN_ONE_DEG_RESOLUTION;
  config->length_resolution = SICK_MEASURE_IN_CM;
  config->baud = SICK_9600_BAUD;
#else
  config->control_mode = SICK_SCAN_ON_RQST_MODE;
  config->angle_range = SICK_SCAN_180_DEG;
  config->angle_resolution = SICK_SCAN_HALF_DEG_RESOLUTION;
  config->length_resolution = SICK_MEASURE_IN_MM;
  config->baud = SICK_9600_BAUD;
#endif


  /* init of user config */
#ifdef SICK_NOT_CONTROLLABLE
  SDI_F->required_control_mode = SICK_SCAN_CONTINUOUSLY_MODE;
  user->angle_range = SICK_SCAN_180_DEG;
  user->angle_resolution = SICK_SCAN_ONE_DEG_RESOLUTION;
  user->length_resolution = SICK_MEASURE_IN_CM;
  user->baud = SICK_9600_BAUD;
#else
  SDI_F->required_control_mode = SICK_SCAN_CONTINUOUSLY_MODE;
  user->angle_range = SICK_SCAN_180_DEG;
  user->angle_resolution = SICK_SCAN_HALF_DEG_RESOLUTION;
  user->length_resolution = SICK_MEASURE_IN_MM;
  user->baud = SICK_38400_BAUD;
#endif


  /* init poster acquisition */
  m->header.np = 0;
  m->header.theta_min = 0;
  m->header.dtheta = 0;


  /* Parameters for segmentation */
  SDI_F->segParams.leastSquareThres = SICK_DEFAULT_SEG_LEAST_SQUARE_THRES;
  SDI_F->segParams.nbMinSegPts = SICK_DEFAULT_SEG_NB_MIN_PTS;
  SDI_F->segParams.distMaxSegPts = SICK_DEFAULT_SEG_DIST_MAX_PTS;

  SDI_F->initParams.acqWay = SICK_ACQ_REAL;
  SDI_F->initParams.exportData.exportPolarPoints = SICK_TRUE;
  SDI_F->initParams.exportData.exportCartesianPoints = SICK_FALSE;
  SDI_F->initParams.exportData.exportSegments = SICK_FALSE;
  SDI_F->initParams.exportData.cartesianFrame = SICK_GLOBAL_FRAME;

/*
  strcpy(SDI_F->initParams.serialPort.name, SICK_SERIAL_DEVICE_NAME);
  */

  /* Read Sick data by file */
  strcpy(SDI_F->saveFileParams.baseName, SICK_DEFAULT_SICK_DATA_BASE_NAME);
  SDI_F->saveFileParams.num=SICK_DEFAULT_SICK_DATA_FILE_NUM;
  
}

/* ----------------------------------------------------------------------
 * is measure tram
 */
BOOL
sick_wait_measures_tram (int *bilan) 
{
#ifdef SICK_UPSIDE_DOWN
  double d [SICK_N_PT_MAX];
#else
  double *d = SDI_F->polarMeasures.measures.rho;
#endif
  double *theta = SDI_F->polarMeasures.theta;
  int size;
  BOOL bilan_local = FALSE;
  static int index = 0;
  static POM_SENSOR_POS pos[SICK_N_ROBOT_POS];
  static int timesnochange;

  /* robot pos */
  if (sick_get_robot_position (pos+index,bilan) != OK) {
    *bilan = S_sick_cannot_connect_to_pom;
    fprintf(stderr,"SICK: cannot read pos\n");
/*     return FALSE; */
  }
  
  /* check trams */
  size = sick_lib_read_last_entry_tram (sick_id);

  if (size) {
    if (sick_lib_is_measure_tram (sick_id)) {

      /* Update sick pos (for simulation) */
      sick_update_global_sick_pos(&pos[index], sick_id);

      /* update internal str */
      if (sick_lib_update_sick_str (sick_id, d, theta) != OK)
	fprintf (stderr, "\t->Error in Sick Logic!\n");

      /* sick is upside down : change order */
#ifdef SICK_UPSIDE_DOWN
      {
	double *rhoP, *dP;
	int i;
	rhoP = &SDI_F->polarMeasures.measures.rho[0];
	dP = &d[SDI_F->polarMeasures.measures.header.np];
	for (i=0;i<SDI_F->polarMeasures.measures.header.np;i++) {
	  *rhoP = *dP;
	  rhoP++;
	  dP--;
	}
      }
#endif
#ifdef SICK_MIN_ANGLE
      {
	int i=0;
	double angle = SDI_F->polarMeasures.measures.header.theta_min;
	while (angle < SICK_MIN_ANGLE) {
	  SDI_F->polarMeasures.measures.rho[i] = SDI_F->sick_config.length_range_max;
	  i++;
	  angle += SDI_F->polarMeasures.measures.header.dtheta;
	}
      }
#endif
#ifdef SICK_MAX_ANGLE
      {
	int i=SDI_F->polarMeasures.measures.header.np-1;
	double angle = SDI_F->polarMeasures.measures.header.theta_min +
		       SDI_F->polarMeasures.measures.header.dtheta * 
		       SDI_F->polarMeasures.measures.header.np;
	while (angle > SICK_MAX_ANGLE) {
	  SDI_F->polarMeasures.measures.rho[i] = SDI_F->sick_config.length_range_max;
	  i--;
	  angle -= SDI_F->polarMeasures.measures.header.dtheta;
	}
      }
#endif
      /* choose robot pos */
      sick_choose_robot_pos (index, pos);

      /* update time */
      SDI_F->polarMeasures.measures.header.send_time = sick_id->state.start;
      SDI_F->polarMeasures.measures.header.rcv_time = sick_id->state.end;
     
      /* save start time */
      if (sick_id->state.start == sick_id->state.end) {
	printf ("sick : warning date does not change for %d times\n", 
		timesnochange);
	timesnochange++;
      }
      else {
	sick_id->state.start = sick_id->state.end;
	timesnochange = 0;
      }

      /* ok */
      bilan_local =  TRUE;
    } else {

      logMsg ("Warnning: Measures TGM Was Expected\n");
      bilan_local =  FALSE;
    }
  } 

  /* increment index */
  index++;
  if (index >= SICK_N_ROBOT_POS) {
    /* scroll */
    memmove ((void *)pos, &pos[1], sizeof (POM_SENSOR_POS)*(SICK_N_ROBOT_POS - 1));
    index--;
  }

  return bilan_local;
}


/* ----------------------------------------------------------------------
 * Ask for measures
 */
STATUS
sick_ask_for_measures (int *bilan)
{
  if (sick_lib_get_measures_cmd (sick_id) != OK) {
    *bilan = errnoGet ();
    return ERROR;
  }

  return OK;
}

/* ----------------------------------------------------------------------
 * Sick dynamic configurations
 */
ACTIVITY_EVENT
sick_dynamic_configurations (int *bilan)
{
#ifdef SICK_NOT_CONTROLLABLE
  SDI_F->state = SICK_CONFIGURATED_STATE;
  return ETHER;
#else
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;
  int size;
  
  switch (parser->state) {
  case SICK_FREE_STATE:
    break;
  case SICK_WAITING_RESPONSE_STATE:
    /* check trams */
    size = sick_lib_read_entry_tram (sick_id);

    if (size) {
      printf ("sick_dynamic_configurations read size %d\n", size);
      /* show tram */
      sick_lib_tram_show (&sick_id->tram);
	
      /* update internal str */
      if (sick_lib_update_sick_str (sick_id, 0, 0) != OK)
	fprintf (stderr, "\t->Error in Sick Logic!\n");
	
      /* is it the waited response */
      if (sick_lib_is_good_ack_cmd (sick_id)) {
/* 	printf ("\t\t-> Good TGM!\n"); */
	/* update internal str */
	sick_update_sick_data_str (parser->last);

	if (parser->last == SICK_SET_BAUD_RATE_STATE) {
	  if (sick_serial_change_baud(sick_id->rcv.baud, bilan) != OK)
	    return FAIL;
	}
	return END;
      } else {
	/* Special case of reset Ack */
	if (parser->last == SICK_RESET_STATE) {
	  /* reset to default rate */
	  if (sick_serial_change_baud(SICK_SERIAL_DEFAULT_RATE, bilan) != OK)
	    return FAIL;
	} else {
	  *bilan = S_sick_bad_tram;
	  printf ("\t\t-> Bad TGM!\n");
	  return FAIL;
	}
      }
    } else {
      /* exec */
      return EXEC;
    }
    break;
  case SICK_RESET_STATE:
    /* reset sick */
    if (sick_lib_reset_cmd (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_READ_STATUS_STATE:
    /* read sick status  */
    if (sick_lib_read_state_cmd (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_READ_CONFIG_STATE:
    /* read sick config  */
    if (sick_lib_read_config_cmd (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_CONFIG_MODE_STATE:
    /* put sick in config mode  */
    if (sick_lib_config_mode (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_SET_CONFIG_STATE:
    break;
  case SICK_SET_MEASURE_UNIT_STATE:
    /* set length_resolution */
    sick_set_length_resolution ();
    /* send to sick  */
    if (sick_lib_set_measure_unit (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_SET_SCANNING_PARAMETERS_STATE:
    /* set scanning parameters */
    sick_set_scanning_params ();
    /* send to sick  */
    if (sick_lib_set_scan_params_cmd (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_SET_BAUD_RATE_STATE:
    /* set baud rate */
    sick_set_baud_rate ();
    /* send to sick  */
    if (sick_lib_set_baud_rate_cmd (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_SET_SCAN_ON_RQST_STATE:
    /* send to sick  */
    if (sick_lib_one_cut_mode (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_SET_SCAN_CONT_STATE:
    /* send to sick  */
    if (sick_lib_cont_cut_mode (sick_id) != OK) {
      *bilan = S_sick_serial_output_error;
      return FAIL;
    }
    break;
  case SICK_CONFIGURATED_STATE:
    SDI_F->state = SICK_CONFIGURATED_STATE;
    SDI_F->control_mode = SDI_F->sick_config.control_mode;
    return ETHER;
  } /* switch */

  /* set last state */
  parser->last = parser->state;
  parser->state = SICK_WAITING_RESPONSE_STATE;

  return EXEC;

#endif /* SICK_NOT_CONTROLLABLE */
}


/* ----------------------------------------------------------------------
 * Sick configurate 
 */
static STATUS
sick_first_configurations (int *bilan)
{
#ifdef SICK_NOT_CONTROLLABLE
  SDI_F->state = SICK_CONFIGURATED_STATE;
  return OK;
#else
  SICK_SCAN_TASK_STATE state = SICK_RESET_STATE, last_state, next_state;
  int size, t = 0;
  int delay = 5;

  next_state = last_state = state;

  while (1) {
    switch (state) {
    case SICK_FREE_STATE:
      break;
    case SICK_WAITING_RESPONSE_STATE:
      /* check trams */
      size = sick_lib_read_entry_tram (sick_id);
      t += delay;

      if (size) {

#ifdef DEBUG
	printf ("sick_first_configurations read size %d\n", size);
#endif
	/* show tram */
	sick_lib_tram_show (&sick_id->tram);
	
	/* update internal str */
	if (sick_lib_update_sick_str (sick_id, 0, 0) != OK)
	  fprintf (stderr, "\t->Error in Sick Logic!\n");
	
	/* is it the waited response */
	if (sick_lib_is_good_ack_cmd (sick_id)) {
/* 	  printf ("\t\t-> Good TGM!\n"); */
	  t = 0;
	  state = next_state;
	  /* update internal str */
	  sick_update_sick_data_str (last_state);
	  if (last_state == SICK_SET_BAUD_RATE_STATE) {
	    if (sick_serial_change_baud(sick_id->rcv.baud, bilan) != OK)
	      return ERROR;
	  }
	  break;
	} else {
	  /* Special case of reset Ack */
	  if (last_state == SICK_RESET_STATE) {
	    /* reset to default rate */
	    if (sick_serial_change_baud(SICK_SERIAL_DEFAULT_RATE, bilan) != OK)
	      return FAIL;
	  } else {
	    *bilan = S_sick_bad_tram;
	    printf ("\t\t-> Bad TGM!\n");
	  }
	}
      } else {
	if ((t % 200) == 0)
	  printf ("\t\t-> Waiting Response!\n");
      }
      /* check time */
      if (t > SICK_REPONSE_TIME_MAX) {
	printf ("\t\t-> Time out!\n");
	*bilan = S_sick_time_out;
	return ERROR;
      }

      break;
    case SICK_RESET_STATE:
      /* flush in buffer */
      if (sick_flush_serial (bilan) != OK) {
	return ERROR;
      }
      /* reset sick */
      if (sick_lib_reset_cmd (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_READ_STATUS_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
      break;
    case SICK_READ_STATUS_STATE:
      /* read sick status  */
      if (sick_lib_read_state_cmd (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_READ_CONFIG_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
      break;
    case SICK_READ_CONFIG_STATE:
      /* read sick config  */
      if (sick_lib_read_config_cmd (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_CONFIG_MODE_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
      break;
    case SICK_CONFIG_MODE_STATE:
      /* put sick in config mode  */
      if (sick_lib_config_mode (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_SET_MEASURE_UNIT_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
      break;
    case SICK_SET_CONFIG_STATE:
      break;
    case SICK_SET_MEASURE_UNIT_STATE:
      /* set length_resolution */
      sick_set_length_resolution ();
      /* send to sick  */
      if (sick_lib_set_measure_unit (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_SET_SCANNING_PARAMETERS_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
 
      break;
    case SICK_SET_SCANNING_PARAMETERS_STATE:
      /* set scanning parameters */
      sick_set_scanning_params ();
      /* send to sick  */
      if (sick_lib_set_scan_params_cmd (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_SET_BAUD_RATE_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;

      break;
    case SICK_SET_BAUD_RATE_STATE:
      /* set baud rate */
      sick_set_baud_rate ();
      /* send to sick  */
      if (sick_lib_set_baud_rate_cmd (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      if (SDI_F->required_control_mode == SICK_SCAN_ON_RQST_MODE)
	next_state = SICK_SET_SCAN_ON_RQST_STATE;
      else
	next_state = SICK_SET_SCAN_CONT_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
     
      break;
    case SICK_SET_SCAN_ON_RQST_STATE:
      /* send to sick  */
      if (sick_lib_one_cut_mode (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_CONFIGURATED_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
     
      break;
    case SICK_SET_SCAN_CONT_STATE:
      /* send to sick  */
      if (sick_lib_cont_cut_mode (sick_id) != OK) {
	*bilan = S_sick_serial_output_error;
	return ERROR;
      }
      /* set next and last state */
      next_state = SICK_CONFIGURATED_STATE;
      last_state = state;
      state = SICK_WAITING_RESPONSE_STATE;
      t = 0;
      break;
    case SICK_CONFIGURATED_STATE:
      SDI_F->state = SICK_CONFIGURATED_STATE;
      SDI_F->control_mode = SDI_F->sick_config.control_mode;
      return OK;
    } /* switch */

    taskDelay (delay);
  } /* while */
#endif /* SICK_NOT_CONTROLLABLE */
}


/* ----------------------------------------------------------------------
 * Set sick length_resolution
 */
static void
sick_set_length_resolution ()
{
  /* set measures unit */
  switch (SDI_F->user_config.length_resolution) {
  case SICK_MEASURE_IN_MM:
    sick_id->send.measure_unit = SICK_LIB_MEASURE_UNIT_MM;
    break;
  case SICK_MEASURE_IN_CM:
    sick_id->send.measure_unit = SICK_LIB_MEASURE_UNIT_CM;
    break;
  case SICK_MEASURE_IN_10CM:
    sick_id->send.measure_unit = SICK_LIB_MEASURE_UNIT_10_CM;
  } /* switch */
}

/* ----------------------------------------------------------------------
 * set sick scanning params
 */

static void
sick_set_scanning_params ()
{
  /* set scan range */
  switch (SDI_F->user_config.angle_range) {
  case SICK_SCAN_180_DEG:
    sick_id->send.range = SICK_LIB_SCAN_RANGE_180;
    break;
  case SICK_SCAN_100_DEG:
    sick_id->send.range = SICK_LIB_SCAN_RANGE_100;
  } /* switch */

  /* set scan angle_resolution */
  switch (SDI_F->user_config.angle_resolution) {
  case SICK_SCAN_ONE_DEG_RESOLUTION:
    sick_id->send.resolution = SICK_LIB_SCAN_RESOLUTION_100;
    break;
  case SICK_SCAN_HALF_DEG_RESOLUTION:
    sick_id->send.resolution = SICK_LIB_SCAN_RESOLUTION_50;
  } /* switch */
}


/* ----------------------------------------------------------------------
 * Set sick baud rate
 */
static void
sick_set_baud_rate ()
{
  /* set baud rate */
  switch (SDI_F->user_config.baud) {
  case SICK_9600_BAUD:
    sick_id->send.baud = 9600;
    break;
  case SICK_19200_BAUD:
    sick_id->send.baud = 19200;
    break;
  case SICK_38400_BAUD:
    sick_id->send.baud = 38400;
    break;
  } /* switch */
}


/* ----------------------------------------------------------------------
 * Update user <-> sick config
 */
static void
sick_update_sick_data_str (SICK_SCAN_TASK_STATE state)
{
  switch (state) {
  case SICK_FREE_STATE:
    break;
  case SICK_WAITING_RESPONSE_STATE:
      break;
  case SICK_RESET_STATE:
    break;
  case SICK_READ_STATUS_STATE:
    break;
  case SICK_READ_CONFIG_STATE:
    break;
  case SICK_CONFIG_MODE_STATE:
    break;
  case SICK_SET_CONFIG_STATE:
    break;
  case SICK_SET_MEASURE_UNIT_STATE:
    SDI_F->sick_config.length_resolution = SDI_F->user_config.length_resolution; 
    SDI_F->sick_config.length_range_max = sick_lib_get_length_range(sick_id);
    break;
  case SICK_SET_SCANNING_PARAMETERS_STATE:
    SDI_F->sick_config.angle_range = SDI_F->user_config.angle_range; 
    SDI_F->sick_config.angle_resolution = SDI_F->user_config.angle_resolution; 
    break;
  case SICK_SET_BAUD_RATE_STATE:
    SDI_F->sick_config.baud = SDI_F->user_config.baud; 
    break;
  case SICK_SET_SCAN_ON_RQST_STATE:
  case SICK_SET_SCAN_CONT_STATE:
    SDI_F->sick_config.control_mode = SDI_F->required_control_mode;
    break;
  default:
    return;
  } /* switch */ 
}


/* ----------------------------------------------------------------------
 * change serial baud rate
 */
static STATUS
sick_serial_change_baud (int baud, int *bilan)
{
  /* call to serial package */
  if (serial_change_baud_rate (sick_id->serial, baud) != OK) {
    *bilan = errnoGet ();
    return ERROR;
  }

  return OK;
}


/* ----------------------------------------------------------------------
 * Generate parser command
 */
int
sick_compose_parser_cmds_list (int reset)
{
  /* get pointer on parser control */
  SICK_PARSER_CONTROL_STR *parser = &SDI_F->parser;
  /* get pointer of sick and user config str */
  SICK_CONFIG_STR *config = &SDI_F->sick_config;
  SICK_USER_STR *user = &SDI_F->user_config;
  int n = 0;

  /* Init */
  parser->ncmd = 0;
  parser->index = 0;

  /* reset */
  if (reset) {
    n = SICK_RESET_CMDS_LIST_NUMBER;
    memcpy (parser->cmds, reset_cmds, 
	    sizeof (SICK_SCAN_TASK_STATE) * n);
    parser->ncmd = n;
    if (SDI_F->required_control_mode == SICK_SCAN_ON_RQST_MODE)
      parser->cmds[n] = SICK_SET_SCAN_ON_RQST_STATE;
    else
      parser->cmds[n] = SICK_SET_SCAN_CONT_STATE;

    parser->ncmd++;
    parser->index = 0;
    return TRUE;
  }
  
  /* baud */
  if (config->baud != user->baud || config->length_resolution != user->length_resolution) {
    parser->cmds[n++] = SICK_CONFIG_MODE_STATE;

    if (config->length_resolution != user->length_resolution)
      parser->cmds[n++] = SICK_SET_MEASURE_UNIT_STATE;

    if (config->baud != user->baud)
      parser->cmds[n++] = SICK_SET_BAUD_RATE_STATE;
  }

  /* change scan params */
  if (config->angle_range != user->angle_range || 
      config->angle_resolution != user->angle_resolution) 
    parser->cmds[n++] = SICK_SET_SCANNING_PARAMETERS_STATE;

  /* nothing to be done */
  if (config->control_mode == SDI_F->required_control_mode && n == 0)
    return FALSE;

  /* re-set mode */
  if (SDI_F->required_control_mode == SICK_SCAN_ON_RQST_MODE)
    parser->cmds[n++] = SICK_SET_SCAN_ON_RQST_STATE;
  else
    parser->cmds[n++] = SICK_SET_SCAN_CONT_STATE;

    parser->ncmd = n;
    parser->index = 0;

    return TRUE;
}

/* ----------------------------------------------------------------------
 * Find baud rate -- Blocked function
 * This function try to send Get Type Command Using all possible rates
 */
int 
sick_find_baud_rate (SICK_BAUD_RATE *br, int *bilan)
{
#ifdef SICK_NOT_CONTROLLABLE
  return SICK_SERIAL_DEFAULT_RATE;
#else
  int baud[3] = {9600, 19200, 38400};
  SICK_BAUD_RATE rate[3] = {SICK_9600_BAUD, SICK_19200_BAUD, SICK_38400_BAUD};
  int n = 0;
  int time;

  /* print */
  printf ("\n\t -> Finding Baud Rate: ");

  /* init of type */
  sick_id->state.type[0] = 0;

  for ( n = 0; n < 3; n++) {
    /* change serial baud */
    if (serial_change_baud_rate (sick_id->serial, baud[n]) != OK) {
      *bilan = errnoGet ();
      h2printErrno (*bilan);
      return -1;
    }

    /* set wait time */
    time = sysClkRateGet() * 2;
    
    /* flush in buffer */
    if (sick_flush_serial (bilan) != OK) {
      return -1;
    }

    /* send get type command */
    if (sick_lib_get_type_cmd (sick_id) != OK) {
       *bilan = errnoGet ();
      h2printErrno (*bilan);
      return -1;
    }

    /* Wait response with time out */
    while (--time) {
      int size;

      /* check trams */
      size = sick_lib_read_entry_tram (sick_id);

      if (size) {

/* 	printf ("sick_find_baud_rate read size %d\n", size); */

	/* show tram */
	sick_lib_tram_show (&sick_id->tram);

	/* update internal str */
	if (sick_lib_update_sick_str (sick_id, 0, 0) != OK)
	  fprintf (stderr, "Error in Sick Logic!\n");
	
	/* is it the waited response */
	if (!sick_lib_is_good_ack_cmd (sick_id))
	  printf ("Bad TGM -- ");

	/* flush in buffer */
	if (sick_flush_serial (bilan) != OK) {
	  return -1;
	}
	printf (" %d -- OK.\n", baud[n]);
	if (br)
	  *br = rate[n];

	/* return baud */
	return baud[n];
      }

      /* wait 5 msec */
      taskDelay (1);
    } /* while */

  } /* for */

  puts ("Error.");

  
  /* reset serial baud */
  if (serial_change_baud_rate (sick_id->serial, baud[0]) != OK) {
      *bilan = errnoGet ();
      h2printErrno (*bilan);
      return -1;
  }
  
  *bilan = S_sick_cannot_find_baud_rate;
  return -1;
#endif /* SICK_NOT_CONTROLLABLE */

}


/* ----------------------------------------------------------------------
 * sick get robot  position
 */
static STATUS
sick_get_robot_position (POM_SENSOR_POS *pos, int *bilan)
{
  static POM_POS pomPos;
  static POM_ME_POS pomMePos;
  STATUS status;

  /* no poster to read */
  if (SDI_F->initParams.sickPosPosterName.name[0] == '\0') {
    pos->sensorToMain.euler = SDI_F->sickPosOnRobot;
    return OK;
  }

  /* Try to find out one */
  if (!sickPosPosterId) {
    if (!sickPosterFind()) {
      printf ("sick_get_robot_position:  cannot find pos poster %s\n",    
	      SDI_F->initParams.sickPosPosterName.name);
      return ERROR;
    }
  }
    
  /* Get the current position estimate */
  switch(sickPosPosterType) {
  case SICK_POM_SENSOR_POS_POSTER:
#ifdef SICK2
    status = sick2POM_SENSOR_POSPosterRead(sickPosPosterId, pos);
#else
    status = sickPOM_SENSOR_POSPosterRead(sickPosPosterId, pos);
#endif
    return OK;

  case SICK_POM_POS_POSTER:
#ifdef SICK2
    status = sick2POM_POSPosterRead(sickPosPosterId, &pomPos);
#else
    status = sickPOM_POSPosterRead(sickPosPosterId, &pomPos);
#endif
    if (status == OK) {
      pos->date = pomPos.date;
      pos->mainToBase = pomPos.mainToBase;
      pos->mainToOrigin = pomPos.mainToOrigin;
      pos->VLocal = pomPos.VLocal;
      pos->sensorToMain.euler = SDI_F->sickPosOnRobot;
      return OK;
    }
    break;


  case SICK_POM_MEPOS_POSTER:
#ifdef SICK2
    status = sick2POM_ME_POSPosterRead(sickPosPosterId, &pomMePos);
#else
    status = sickPOM_ME_POSPosterRead(sickPosPosterId, &pomMePos);
#endif
    if (status == OK) {
/*       if (pomMePos.kind == POM_ME_DELTA) { */
/* 	printf ("sick: cannot use relative POM_ME_POS\n"); */
/* 	return ERROR; */
/*       } */
      pos->date = pomMePos.date1;
      pos->mainToBase = pomMePos.main;
      pos->mainToOrigin = pomMePos.main;
      pos->VLocal = pomMePos.VLocal;
      pos->sensorToMain.euler = SDI_F->sickPosOnRobot;
      return OK;
    }
    break;


  case SICK_NO_POS_POSTER:
    printf ("sick: no poster ?\n");
    break;
  }

  if (status == ERROR) {
      printf ("sickReadRobotPosition:  cannot read poster %s\n",
	      SDI_F->initParams.sickPosPosterName.name);
      sickPosPosterId = NULL;
      return ERROR;
    }

  /* ok */
  return OK;
}

/* ----------------------------------------------------------------------
 * sick update measures in posters
 */

STATUS
sick_update_measures (int *bilan)
{
  /* get pointer on measures str */
  SICK_MEASURES_STR *m = &SDI_F->polarMeasures.measures;
  SICK_CART_MEASURES_STR *cm;

  int np = m->header.np;

  if (SDI_F->initParams.acqWay != SICK_ACQ_REPLAY_FILES) {
    /* number */
    m->header.np = sick_id->measures.n;
    /* theta min in sick frame */
    m->header.theta_min = sick_id->measures.theta_min;
    /* dtheta */
    m->header.dtheta = sick_id->measures.dtheta;
  }

  /* frame */
  m->header.frame = SICK_SICK_FRAME;

  /**
   **  Polar points
   **/ 

  /* Polar points not exported */
  if (!SDI_F->initParams.exportData.exportPolarPoints && np != 0) {

    /* number */
    m->header.np = 0;

    /* Write header poster */
    if (sickScanPolarHeaderPosterWrite (SICK_SCANPOLAR_POSTER_ID, 
					&(m->header)) == ERROR) { 
      *bilan = errnoGet ();
      return ERROR;
    }

    m->header.np = np;
  }
   
  /* Polar points exported */
  else {

    /* Write all poster */
    if (sickScanPolarPosterWrite (SICK_SCANPOLAR_POSTER_ID, 
				  (SICK_SCANPOLAR_POSTER_STR *)m) 
	== ERROR) { 
      *bilan = errnoGet ();
      return ERROR;
    }
  }  

  /**
   **  Cartesian points
   **/ 
  cm = &SDI_F->cartMeasures;
  cm->header = m->header;

  /* Cartesian points not exported */
  if (!SDI_F->initParams.exportData.exportCartesianPoints) {

    /* number */
    cm->header.np = 0;

    /* Write header poster */
    if (sickScanCartHeaderPosterWrite (SICK_SCANCART_POSTER_ID, 
				       &(cm->header)) == ERROR) { 
      *bilan = errnoGet ();
      return ERROR;
    }
  }

  /* Polar points exported */
  else {

    /* cartesian points in global frame */
    sickPolarToCart1 (SDI_F->polarMeasures.measures.header.np,
		      &SDI_F->polarMeasures.measures.header.sickPomPos,
		      SDI_F->initParams.exportData.cartesianFrame,
		      SDI_F->polarMeasures.measures.rho,
		      SDI_F->polarMeasures.theta,
		      SDI_F->cartMeasures.cartesian_pt);
    SDI_F->cartMeasures.header.frame = SDI_F->initParams.exportData.cartesianFrame;

    /* Write all poster */
    if (sickScanCartPosterWrite (SICK_SCANCART_POSTER_ID, 
				 (SICK_SCANCART_POSTER_STR *) cm) == ERROR) { 
      *bilan = errnoGet ();
      return ERROR;
    }
  }

  /* OK */
  return OK;
}

/* ----------------------------------------------------------------------
 * flush serial port
 */
STATUS
sick_flush_serial (int *bilan)
{
  /* flush in buffer */
  if (serial_flush_in_buffer (sick_id->serial) != OK) {
    *bilan = errnoGet ();
    h2printErrno (*bilan);
    return ERROR;
  }
  return OK;
}


/* ----------------------------------------------------------------------
 * choose robot pos
 */
#define NB_BAUDS_PAR_OCTET 10.


static void
sick_choose_robot_pos (int n, POM_SENSOR_POS *p)
{
  double t;
  int index;

  /* transmission time for the last TGM */
  /* (number of task cycle, ie of record position, during acquisition) */
  t = .5 + (sick_id->tram.len * NB_BAUDS_PAR_OCTET 
	    / (double) sick_id->send.baud 
	    / EXEC_TASK_PERIOD(SICK_SCANTASK_NUM));

  /* number of tick */
  index = t;
  index = n - index + SDI_F->synchroPosTunning;
  if (index > n) index = n - 1;
  if (index < 0) index = 0;

  /* set position */
  memcpy(&SDI_F->polarMeasures.measures.header.sickPomPos,&p[index],sizeof(POM_SENSOR_POS));

}

/** ----------------------------------------------------------------------
 ** sick_update_global_sick_pos  - computes global position of the sick
 **                            for simulation purpose only
 **/

static void sick_update_global_sick_pos(POM_SENSOR_POS *robot,
					SICK_LIB_STR *sick_id)
{
  double x, y, theta;
  /* double phi; */
  double rsin, rcos;
#ifdef READ_SIMULATED_POS
  GENPOS_CART_STATE realRobotPos; /* robot pos in global frame */
#endif

  /* Sick pos on robot */
  x =  robot->sensorToMain.euler.x;
  y =  robot->sensorToMain.euler.y;
  theta = robot->sensorToMain.euler.yaw;


  /* angle robot */  
  
  rsin = sin (robot->mainToBase.euler.yaw) ;
  rcos = cos (robot->mainToBase.euler.yaw) ;

#ifdef READ_SIMULATED_POS
  /* when this flag is set, sick and loco are simulated
     It means sick must read realRobot position and not pom position.
     This realRobot position is produced by loco as a not `noised' odo position.
     See file loco/modifLoco for details. */

  /* read loco poster */
  locoRobotRealRobotPosterRead(&realRobotPos);

  /* compute sick pos in global frame */
  rsin = sin (realRobotPos.theta) ;
  rcos = cos (realRobotPos.theta) ;
  sick_id->x = realRobotPos.xRob + x * rcos - y * rsin;
  sick_id->y = realRobotPos.yRob + x * rsin + y * rcos;
  sick_id->theta = angleLimit(realRobotPos.theta + theta);
#else
  /* Sick pos in global frame */
  sick_id->x = robot->mainToOrigin.euler.x + x * rcos - y * rsin;
  sick_id->y = robot->mainToOrigin.euler.y + x * rsin + y * rcos;
  sick_id->theta = angleLimit(robot->mainToOrigin.euler.yaw + theta);
#endif
}




/* ----------------------------------------------------------------------
 * read environnement for simulation
 */
BOOL sick_func_read_file_env(char *file)
{
  if (sick_simul_read_env(file) == ERROR) {
    printf ("sick_func_read_file_env: read env %s failed\n", file);
    return FALSE;
  }
  return TRUE;
}


/* ----------------------------------------------------------------------
 *  read sick data saved in a file
 */
BOOL sick_read_measure_in_file()
{
  char fileName[64];
  sprintf(fileName, "%s.%05d", 
	  SDI_F->saveFileParams.baseName, SDI_F->saveFileParams.num);  
  printf ("read %s\n", fileName);
  SDI_F->saveFileParams.num++;
  return sickReadFile(fileName, &SDI_F->sick_config, 
		      &SDI_F->polarMeasures.measures.header, SDI_F->polarMeasures.measures.rho);
}

